package es.uoc.mxpdroid.algorithm;

import es.uoc.mxpdroid.entity.RasterLayer;


public class SlopeAlgorithm {

   public final static int    METHOD_MAXIMUM_SLOPE = 0;
   public final static int    METHOD_TARBOTON      = 1;
   public final static int    METHOD_BURGESS       = 2;
   public final static int    METHOD_BAUER         = 3;
   public final static int    METHOD_HEERDEGEN     = 4;
   public final static int    METHOD_ZEVENBERGEN   = 5;
   public final static int    METHOD_HARALICK      = 6;

   public final static int    UNITS_RADIANS        = 0;
   public final static int    UNITS_DEGREES        = 1;
   public final static int    UNITS_PERCENTAGE     = 2;

//   public static final String METHOD               = "METHOD";
//   public static final String UNITS                = "UNITS";
//   public static final String DEM                  = "DEM";
//   public static final String SLOPE                = "SLOPE";

   private final static int   m_iOffsetX[]         = { 0, 1, 1, 1, 0, -1, -1, -1 };
   private final static int   m_iOffsetY[]         = { 1, 1, 0, -1, -1, -1, 0, 1 };

   private RasterLayer  m_DEM;
   private RasterLayer  m_Slope;
   private int          m_iUnits;
   private int			m_Method;
   private double       _2DX;
   private double       _6DX;


   public SlopeAlgorithm( RasterLayer dem, int method, int units) {
	   m_DEM = dem;
	   m_iUnits = units;
	   m_Method = method;
	   
	   processAlgorithm();
   }
   
   public RasterLayer getResultLayer(){
       return m_Slope;
   }
   
   private void processAlgorithm() {

      int x, y;
      int iNX, iNY;

      m_Slope = new RasterLayer(m_DEM.getLayerType());
      m_Slope.initFromLayer(m_DEM);
      
      iNX = m_DEM.getNX();
      iNY = m_DEM.getNY();

      _2DX = m_DEM.getWindowCellSize( (int)iNX/2, (int)iNY ) * 2.;
      _6DX = m_DEM.getWindowCellSize( (int)iNX/2, (int)iNY ) * 6.;
      
      //Aqui tenemos que tener cuidado para no salirnos de la ventana
      for (y = 0; y < iNY; y++) {
         for (x = 0; x < iNX; x++) {
            switch (m_Method) {
               case 0:
                  Do_MaximumSlope(x, y);
                  break;

               case 1:
                  Do_Tarboton(x, y);
                  break;

               case 2:
                  Do_LeastSquare(x, y);
                  break;

               case 3:
                  Do_FD_BRM(x, y);
                  break;

               case 4:
                  Do_FD_Heerdegen(x, y);
                  break;

               case 5:
                  Do_FD_Zevenbergen(x, y);
                  break;

               case 6:
                  Do_FD_Haralick(x, y);
                  break;
            }
         }
      }

   }


   private void Do_MaximumSlope(final int x,
                                final int y) {

      int i;
      double z, z2, dSlope, dMaxSlope;

      z = m_DEM.getCellValueAsDouble(x, y);

      if (m_DEM.isNoDataValue(z)) {
         Set_Parameters_NoData(x, y);
      }
      else {
         dMaxSlope = 0.0;
         for (i = 0; i < 8; i++) {
            z2 = m_DEM.getCellValueAsDouble(x + m_iOffsetX[i], y + m_iOffsetY[i]);
            if (!m_DEM.isNoDataValue(z2)) {
               dSlope = Math.atan((z - z2) / m_DEM.getDistToNeighborInDir(i));
               if (dSlope > dMaxSlope) {
                  dMaxSlope = dSlope;
               }
            }
         }

         Set_Parameters(x, y, dMaxSlope);
      }
   }


   private void Do_Tarboton(final int x,
                            final int y) {

      int i, j;
      double z, z2, zm[], iSlope, Slope, G, H;

      zm = new double[8];

      z = m_DEM.getCellValueAsDouble(x, y);

      if (m_DEM.isNoDataValue(z)) {
         Set_Parameters_NoData(x, y);
      }
      else {
         for (i = 0; i < 8; i++) {

            z2 = m_DEM.getCellValueAsDouble(x + m_iOffsetX[i], y + m_iOffsetY[i]);
            if (!m_DEM.isNoDataValue(z2)) {
               zm[i] = z2;
            }
            else {
               z2 = m_DEM.getCellValueAsDouble(x + m_iOffsetX[(i + 4) % 8], y + m_iOffsetY[(i + 4) % 8]);
               if (!m_DEM.isNoDataValue(z2)) {
                  zm[i] = z - (z2 - z);
               }
               else {
                  zm[i] = z;
               }
            }
         }

         Slope = 0.0;

         for (i = 0, j = 1; i < 8; i++, j = (j + 1) % 8) {
            if ((i % 2) != 0) // i => diagonal
            {
               G = (z - zm[j]) / m_DEM.getWindowCellSize(x, y);
               H = (zm[j] - zm[i]) / m_DEM.getWindowCellSize(x, y);
            }
            else // i => orthogonal
            {
               G = (z - zm[i]) / m_DEM.getWindowCellSize(x, y);
               H = (zm[i] - zm[j]) / m_DEM.getWindowCellSize(x, y);
            }

            if (H < 0.0) {
               iSlope = G;
            }
            else if (H > G) {
               iSlope = (z - zm[((i % 2) != 0) ? i : j]) / (Math.sqrt(2.0) * m_DEM.getWindowCellSize(x, y));
            }
            else {
               iSlope = Math.sqrt(G * G + H * H);
            }

            if (iSlope > Slope) {
               Slope = iSlope;
            }
         }

         Set_Parameters(x, y, Math.atan(Slope));

      }
   }


   private void Do_LeastSquare(final int x,
                               final int y) {

      double zm[], a, b;

      zm = new double[9];

      if (Get_SubMatrix3x3(x, y, zm)) {
         a = ((zm[2] + 2 * zm[5] + zm[8]) - (zm[0] + 2 * zm[3] + zm[6])) / (8 * m_DEM.getWindowCellSize(x, y));
         b = ((zm[6] + 2 * zm[7] + zm[8]) - (zm[0] + 2 * zm[1] + zm[2])) / (8 * m_DEM.getWindowCellSize(x, y));

         if (a != 0.0) {
            Set_Parameters(x, y, Math.atan(Math.sqrt(a * a + b * b)));
         }
         else if (b > 0.0) {
            Set_Parameters(x, y, Math.atan(Math.sqrt(a * a + b * b)));
         }
         else if (b < 0.0) {
            Set_Parameters(x, y, Math.atan(Math.sqrt(a * a + b * b)));
         }
         else {
            Set_Parameters_NoData(x, y);
         }
      }
   }


   private void Do_FD_BRM(final int x,
                          final int y) {

      double zm[], G, H;

      zm = new double[9];

      if (Get_SubMatrix3x3(x, y, zm)) {
         G = ((zm[2] - zm[0]) + (zm[5] - zm[3]) + (zm[8] - zm[6])) / _6DX;
         H = ((zm[6] - zm[0]) + (zm[7] - zm[1]) + (zm[8] - zm[2])) / _6DX;
         Set_Parameters_Derive(x, y, G, H);
      }
   }


   private void Do_FD_Heerdegen(final int x,
                                final int y) {

      double zm[], G, H;

      zm = new double[9];

      if (Get_SubMatrix3x3(x, y, zm)) {
         G = (-zm[0] + zm[2] - zm[3] + zm[5] - zm[6] + zm[8]) / _6DX;
         H = (-zm[0] - zm[1] - zm[2] + zm[6] + zm[7] + zm[8]) / _6DX;
         Set_Parameters_Derive(x, y, G, H);
      }
   }


   private void Do_FD_Zevenbergen(final int x,
                                  final int y) {

      double zm[], G, H;

      zm = new double[9];

      if (Get_SubMatrix3x3(x, y, zm)) {
         G = (zm[5] - zm[3]) / _2DX;
         H = (zm[7] - zm[1]) / _2DX;
         Set_Parameters_Derive(x, y, G, H);
      }
   }


   private void Do_FD_Haralick(final int x,
                               final int y) {

      final int Mtrx[][][] = {
               { { 31, -5, -17, -5, 31 }, { -44, -62, -68, -62, -44 }, { 0, 0, 0, 0, 0 }, { 44, 62, 68, 62, 44 },
                        { -31, 5, 17, 5, -31 } },
               { { 31, -44, 0, 44, -31 }, { -5, -62, 0, 62, 5 }, { -17, -68, 0, 68, 17 }, { -5, -62, 0, 62, 5 },
                        { 31, -44, 0, 44, -31 } },
               { { 2, 2, 2, 2, 2 }, { -1, -1, -1, -1, -1 }, { -2, -2, -2, -2, -2 }, { -1, -1, -1, -1, -1 }, { 2, 2, 2, 2, 2 } },
               { { 4, 2, 0, -2, -4 }, { 2, 1, 0, -1, -2 }, { 0, 0, 0, 0, 0 }, { -2, -1, 0, 1, 2 }, { -4, -2, 0, 2, 4 } },
               { { 2, -1, -2, -1, 2 }, { 2, -1, -2, -1, 2 }, { 2, -1, -2, -1, 2 }, { 2, -1, -2, -1, 2 }, { 2, -1, -2, -1, 2 } } };

      final int QMtrx[] = { 4200, 4200, 700, 1000, 700 };

      int i, ix, iy, n;
      double Sum, zm[], k[];

      zm = new double[25];
      k = new double[2];

      if (Get_SubMatrix5x5(x, y, zm)) {
         for (i = 0; i < 2; i++) {
            for (n = 0, Sum = 0.0, iy = 0; iy < 5; iy++) {
               for (ix = 0; ix < 5; ix++, n++) {
                  Sum += zm[n] * Mtrx[i][ix][iy];
               }
            }

            k[i] = Sum / QMtrx[i];
         }

         Set_Parameters_Derive(x, y, k[1], k[0]);
      }
   }


   // additional methods

   private void Set_Parameters(final int x,
                               final int y,
                               double dSlope) {

      if (m_iUnits == UNITS_DEGREES) {
         dSlope = Math.toDegrees(dSlope);
      }
      else if (m_iUnits == UNITS_PERCENTAGE) {
         dSlope = Math.tan(dSlope) * 100;
      }

      m_Slope.setCellValue(x, y, dSlope);
      //System.out.println(Integer.toString(x) + "\t" + Integer.toString(x) +"\t" + dSlope);

   }


   private void Set_Parameters_Derive(final int x,
                                      final int y,
                                      final double G,
                                      final double H) {

      double k2;
      k2 = G * G + H * H;

      if (G != 0.0) {
         Set_Parameters(x, y, Math.atan(Math.sqrt(k2)));
      }
      else if (H > 0.0) {
         Set_Parameters(x, y, Math.atan(Math.sqrt(k2)));
      }
      else if (H < 0.0) {
         Set_Parameters(x, y, Math.atan(Math.sqrt(k2)));
      }
      else {
         Set_Parameters(x, y, Math.atan(Math.sqrt(k2)));
      }
   }

   private void Set_Parameters_NoData(final int x,
                                      final int y) {
        m_Slope.setNoDataAtCell(x, y);

   }


   //	---------------------------------------------------------
   //	 Indexing of the Submatrix:
   //
   //	  +-------+    +-------+
   //	  | 7 0 1 |    | 2 5 8 |
   //	  | 6 * 2 | => | 1 4 7 |
   //	  | 5 4 3 |    | 0 3 6 |
   //	  +-------+    +-------+
   //
   //	---------------------------------------------------------
   private boolean Get_SubMatrix3x3(final int x,
                                    final int y,
                                    final double SubMatrix[]) {

      final int iSub[] = { 5, 8, 7, 6, 3, 0, 1, 2 };

      int i;
      double z, z2;

      z = m_DEM.getCellValueAsDouble(x, y);

      if (m_DEM.isNoDataValue(z)) {
         Set_Parameters_NoData(x, y);
      }
      else {
         SubMatrix[4] = 0.0;
         for (i = 0; i < 8; i++) {
            z2 = m_DEM.getCellValueAsDouble(x + m_iOffsetX[i], y + m_iOffsetY[i]);
            if (!m_DEM.isNoDataValue(z2)) {
               SubMatrix[iSub[i]] = z2 - z;
            }
            else {
               z2 = m_DEM.getCellValueAsDouble(x + m_iOffsetX[(i + 4) % 8], y + m_iOffsetY[(i + 4) % 8]);
               if (!m_DEM.isNoDataValue(z2)) {
                  SubMatrix[iSub[i]] = z - z2;
               }
               else {
                  SubMatrix[iSub[i]] = 0.0;
               }
            }
         }

         return (true);
      }

      return (false);
   }


   private boolean Get_SubMatrix5x5(final int x,
                                    final int y,
                                    final double SubMatrix[]) {
      int i, ix, iy, jx, jy;
      double z, z2;

      z = m_DEM.getCellValueAsDouble(x, y);

      if (!m_DEM.isNoDataValue(z)) {
         for (i = 0, iy = y - 2; iy <= y + 2; iy++) {
            jy = iy < 0 ? 0 : (iy >= m_DEM.getNY() ? m_DEM.getNY() - 1 : iy);
            for (ix = x - 2; ix <= x + 2; ix++, i++) {
               jx = ix < 0 ? 0 : (ix >= m_DEM.getNX() ? m_DEM.getNY() - 1 : ix);
               z2 = m_DEM.getCellValueAsDouble(jx, jy);
               if (!m_DEM.isNoDataValue(z2)) {
                  SubMatrix[i] = z2 - z;
               }
               else {
                  SubMatrix[i] = 0.0;
               }
            }
         }

         return (true);
      }

      return (false);
   }

}
