package es.uoc.mxpdroid.entity;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.StringTokenizer;

import es.uoc.mxpdroid.util.GeoUTM;

public class IntRasterLayer {
	
    private final static int   m_iOffsetX[]         = { 0, 1, 1, 1, 0, -1, -1, -1 };
    private final static int   m_iOffsetY[]         = { 1, 1, 0, -1, -1, -1, 0, 1 };
    
	public static final String SRTM		= "SRTM";
	public static final String ICC		= "ICC";
	public static final String ICC30	= "ICC30";
	
	public static final int INTERPOLATION_NearestNeighbour = 0;
	public static final int INTERPOLATION_Bilinear         = 1;
	public static final int INTERPOLATION_InverseDistance  = 2;
	public static final int INTERPOLATION_BicubicSpline    = 3;
	public static final int INTERPOLATION_BSpline          = 4;
	
	public static final double DEG_TO_RAD		= Math.PI / 180.;
	public static final double RAD_TO_DEG		= 180. / Math.PI;
	public final static double DEG_90_IN_RAD 	= Math.PI / 180. * 90.;
	public final static double DEG_180_IN_RAD 	= Math.PI ;
	public final static double DEG_270_IN_RAD 	= Math.PI / 180. * 270.;
	
	public static final int R_EQU				= 6378137;
	public static final double APLA				= 1. / 298.257;
	
    private double  _2DX;
    private double  _6DX;
    
    //number of rows and cols
	private int iNI;
	private int iNJ;
	
	//the matrix containing the data
	private int[][] layer;
	private String layerType;
	private int cellSize;
	
	//lat Max an Lon Min for the SRTM Data
    private double fLatMax;
    private double fLonMin;
    private int m_iMethod;
    
    //x and y in meters to the low-left corner UTM
    private double xll;
    private double yll;
    
    // default no data value for raster layer
    public final static int NO_DATA_VALUE = -32768;
    
	private int no_data = NO_DATA_VALUE;
	
	
	/**
	 * Costructor de la capa raster
	 * @param layer_type nom de la capa
	 */
	public IntRasterLayer(String layer_type){
		layerType = layer_type;
		if (layerType.equals(RasterLayer.ICC))
			this.cellSize = 15;
		else if (layerType.equals(RasterLayer.ICC30))
			this.cellSize = 30;
		_2DX = cellSize * 2.;
        _6DX = cellSize * 6.;
	}
	
	public double getXll(){
		return xll;
	}
	public double getYll(){
		return yll;
	}
	public double getNoDataValue(){
		return this.no_data;
	}
	
	public void printToAsciiFile(File ff) throws Exception{
		FileWriter fw = new FileWriter(ff);
        BufferedWriter bw = new BufferedWriter(fw);
        bw.write("NCOLS\t" + getNY() + "\n");
        bw.write("NROWS\t" + getNX() + "\n");
        bw.write("XLLCORNER\t" + getXll() + "\n");
        bw.write("YLLCORNER\t" + getYll() + "\n");
        bw.write("CELLSIZE\t" + getWindowCellSize(0, 0) + "\n");
        bw.write("NODATA_VALUE\t" + getNoDataValue() + "\n");
        
        for(int kk = 0; kk < layer.length; kk++){
        	for (int jj = 0; jj < layer[kk].length; jj++){
        		bw.write(Float.toString((float)layer[kk][jj]));
        		bw.write("\t");
        	}
        	bw.write("\n");
        }
        
        
        bw.close();
        fw.close();
	}
	
	public void initFromFile(String fileData, File localDir){
		try{
			String iccFileName = fileData;
			File file = new File(iccFileName);
			
			file = new File(localDir, file.getName());
			if (!file.exists()) {
					throw new FileNotFoundException(file.getAbsolutePath() + " not found!");
			}
			FileReader fr = new FileReader(file);
			BufferedReader br = new BufferedReader(fr);
			
			//las 6 primeras linias del fichero nos dans la forma
			StringTokenizer stk;
			final double[] tmpValues = new double[6];
			if (br.ready())
			for (int ii = 0; ii < 6; ii++){
				stk = new StringTokenizer(br.readLine());
				stk.nextToken();
				tmpValues[ii] = Double.parseDouble(stk.nextToken());
			}
			
			//assing constants of layer
			iNJ = (int)tmpValues[0];
			iNI = (int)tmpValues[1];
			xll = tmpValues[2];
			yll = tmpValues[3];
			cellSize = (int) tmpValues[4];
			
			if (cellSize == 15) layerType = RasterLayer.ICC;
			else if (cellSize == 30) layerType = RasterLayer.ICC30;
			_2DX = cellSize * 2.;
	        _6DX = cellSize * 6.;
			
			no_data = (int)tmpValues[5];
			
			layer = new int[iNI][iNJ];
			
			//reading matrix of data
			int ii = 0, jj = 0;
			
			while( br.ready()) {
				stk = new StringTokenizer(br.readLine());
				
				while (stk.hasMoreTokens()){
					layer[ii][jj] = Integer.parseInt(stk.nextToken());
					jj++;
					
				}
				
				if (jj >= iNJ){
					jj=0;
					ii++;
				}
			}
			
			
			br.close();
			fr.close();
			
		}catch(Exception e){
			e.printStackTrace();
		}
	}
	
	/**
	 * This method allows to initialize the raster object from
	 * another raster object in order to keep only the form and not the 
	 * content
	 * @param other other Raster Layer
	 */
	public void initFromLayer(RasterLayer other){
		this.setLayer(other.getLayer()); //init iNX, iNY, _2DX, _6DX and layer
		this.assignNoData(); // set layer to no data values by default NO_DATA_VALUE
		this.setUTMOrig(other.getUTMOrig()); // set xll and yll
		this.cellSize = (int) other.getWindowCellSize((int)(iNI/2), (int)(iNJ/2));
	}
	
	public void setUTMOrig(double[] orig){
		this.xll = (float)orig[0];
		this.yll = (float)orig[1];
	}
	
	public String rasterType(){
		return this.layerType;
	}
	
	public double[] getUTMOrig(){
		double[] tmp = new double[2];
		tmp[0] = xll;
		tmp[1] = yll;
		
		return tmp;
	}
	
	public void setDataAtLatLon(double lat, double lon, int data){
		int iijj[] = get_iijj(lat, lon, GeoUTM.ETRS89, 31);
		layer[iijj[0]][iijj[1]] = data;
	}
	
	public void setDataAtXY_UTM(double xUTM, double yUTM, int data){
		int iijj[] = get_iijj(xUTM, yUTM);
		layer[iijj[0]][iijj[1]] = data;
	}
	
	public int[] get_iijj(double xUTM, double yUTM){
		int[] coords = new int[2];
		
		
//		coords[0] = iNI - (int)((yUTM-yll)/cellSize + 1/2);
//		coords[1] = (int)((xUTM-xll)/cellSize + 1/2);
//		
		coords[0] = iNI - (int)((yUTM-yll)/cellSize) + 1;
		coords[1] = (int)((xUTM-xll)/cellSize);
		
		//here the X coordinate is not good!!! always under zero!!
		
		if (coords[0] < 0) coords[0] = 0;
		if (coords[1] < 0) coords[1] = 0;
		if (coords[0] >= iNI) coords[0] = iNI - 1;
		if (coords[1] >= iNJ) coords[1] = iNJ - 1;
		
		return coords;
	}
	
	public double[] get_latlon(int ii, int jj, int type_ellipsoide, int zoneUTM){
		double[] coords = new double[2];
		double[] xyUTM = new double[2];
		
//		xyUTM[0] = jj * cellSize + xll;
//		xyUTM[1] = (iNI -1 - ii) * cellSize + yll;
		
		//this is the center of grid cell and not the corner
		xyUTM[0] = jj * cellSize + xll + cellSize/2.0;
		xyUTM[1] = (iNI + 1 - ii) * cellSize + yll + cellSize/2.0;
		
		GeoUTM gu = new GeoUTM(type_ellipsoide);
		
		coords = gu.passToGeographic(xyUTM[0], xyUTM[1], 31, false, false);
		
		//coords are in radiant, we need to pass it to degrees
		coords[0] *= RAD_TO_DEG;
		coords[1] *= RAD_TO_DEG;
		
		return coords;
	}
	
	private int[] get_iijj(double lat, double lon, int type_ellipsoide, int zoneUTM){
		int[] coords = new int[2];
		
		GeoUTM gu = new GeoUTM(type_ellipsoide);
		double[] xyUTM = gu.passToUTM(lat, lon, zoneUTM);
		
//		coords[0] = iNI - (int)((xyUTM[1]-yll)/cellSize + 1/2);
//		coords[1] = (int)((xyUTM[0]-xll)/cellSize + 1/2);
		
		coords[0] = iNI - (int)((xyUTM[1]-yll)/cellSize) + 1;
		coords[1] = (int)((xyUTM[0]-xll)/cellSize);
		
		if (coords[0] < 0) coords[0] = 0;
		if (coords[1] < 0) coords[1] = 0;
		if (coords[0] >= iNI) coords[0] = iNI - 1;
		if (coords[1] >= iNJ) coords[1] = iNJ - 1;
		
		return coords;
	}
	
	public String getLayerType(){
		return layerType;
	}
	/**
	 * Initcialitza el buffer
	 * @param b
	 */
	public void setLayer(double[][] bb){
		iNI = bb.length;
	    iNJ = bb[0].length;
		layer = new int[iNI][iNJ];
		for(int ii=0; ii<iNI; ii++)
			for(int jj=0; jj<iNJ; jj++)
				layer[ii][jj] = (int)bb[ii][jj];
       
        _2DX = this.getWindowCellSize( (int)iNI/2, (int)iNJ ) * 2.;
        _6DX = this.getWindowCellSize( (int)iNI/2, (int)iNJ ) * 6.;
	}
	
	public int[][] getLayer(){
		return layer;
	}
	
	public void multiply(double value){
		for (int ii=0; ii<iNI; ii++)
			for(int jj=0; jj<iNJ; jj++)
				if( !isNoDataValue(layer[ii][jj]) )
					layer[ii][jj] *= value;
	}
	
	
	public void assign(int value){
		for (int ii=0; ii<iNI; ii++)
			for(int jj=0; jj<iNJ; jj++)
				layer[ii][jj] = value;
	}
	
	public int getCellValueAsInt(int xx, int yy){
		return (int)layer[xx][yy];
	}
	
	public void setCellSize(int arcseconds){
		cellSize = arcseconds;
	}
	
	public void  setNoDataValue(int value){
		no_data = value;
	}
	
	public void assignNoData(){
		for (int ii=0; ii<iNI; ii++)
			for(int jj=0; jj<iNJ; jj++)
				layer[ii][jj] =  no_data;
	}
	public void setNoDataAtCell(int xx, int yy){
		if (xx>=0 && xx<iNI && yy>=0 && yy<iNJ)
			layer[xx][yy] =  no_data;
			
	}
	
	public double getDistToNeighborInDir(int ii){
		double distance;
		distance = getWindowCellSize((int)(iNI/2), (int)(iNJ/2));
		if (ii%2 != 0) //ii=1, 3, 5 o 7
			distance *= Math.sqrt(2);
		return distance;
	}
	/**
	 * This method allow to 
	 * @param xx
	 * @param yy
	 * @return
	 */
	public double getWindowCellSize(int xx, int yy){
		return cellSize;
	}
	public boolean isNoDataValue(int val){
		if (val == no_data)
			return true;
		return false;
	}
	
	public double getCellValueAsDouble(int xx, int yy){
            
        if( xx >= 0 && xx < iNI && yy >= 0 && yy < iNJ)
              return layer[xx][yy];
        return no_data;
	}
	
	public void setCellValue(int xx, int yy, int value){
		if( xx >= 0 && xx < iNI && yy >= 0 && yy < iNJ)
			layer[xx][yy] = value;
	}
	
	public int getNX(){
		return iNI;
	}
	
	public int getNY(){
		return iNJ;
	}

}
