package es.uoc.mxpdroid.entity;


import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.StringTokenizer;

import es.uoc.mxpdroid.util.GeoUTM;

public class RasterLayer {
	
    private final static int   m_iOffsetX[]         = { 0, 1, 1, 1, 0, -1, -1, -1 };
    private final static int   m_iOffsetY[]         = { 1, 1, 0, -1, -1, -1, 0, 1 };
    
	public static final String SRTM		= "SRTM";
	public static final String ICC		= "ICC";
	public static final String ICC30	= "ICC30";
	
	public static final int INTERPOLATION_NearestNeighbour = 0;
	public static final int INTERPOLATION_Bilinear         = 1;
	public static final int INTERPOLATION_InverseDistance  = 2;
	public static final int INTERPOLATION_BicubicSpline    = 3;
	public static final int INTERPOLATION_BSpline          = 4;
	
	public static final double DEG_TO_RAD		= Math.PI / 180.;
	public static final double RAD_TO_DEG		= 180. / Math.PI;
	public final static double DEG_90_IN_RAD 	= Math.PI / 180. * 90.;
	public final static double DEG_180_IN_RAD 	= Math.PI ;
	public final static double DEG_270_IN_RAD 	= Math.PI / 180. * 270.;
	
	public static final int R_EQU				= 6378137;
	public static final double APLA				= 1. / 298.257;
	
    private double  _2DX;
    private double  _6DX;
    
    //number of rows and cols
	private int iNI;
	private int iNJ;
	
	//the matrix containing the data
	private double[][] layer;
	private String layerType;
	private int cellSize;
	
	//lat Max an Lon Min for the SRTM Data
    private double fLatMax;
    private double fLonMin;
    private int m_iMethod;
    
    //x and y in meters to the low-left corner UTM
    private double xll;
    private double yll;
    
    // default no data value for raster layer
    public final static int NO_DATA_VALUE = -32768;
    
	private double no_data = NO_DATA_VALUE;
	
	
	/**
	 * Costructor de la capa raster
	 * @param layer_type nom de la capa
	 */
	public RasterLayer(String layer_type){
		layerType = layer_type;
		if (layerType.equals(RasterLayer.ICC))
			this.cellSize = 15;
		else if (layerType.equals(RasterLayer.ICC30))
			this.cellSize = 30;
		_2DX = cellSize * 2.;
        _6DX = cellSize * 6.;
	}
	
	public double getXll(){
		return xll;
	}
	public double getYll(){
		return yll;
	}
	public double getNoDataValue(){
		return this.no_data;
	}
	
	public void printToAsciiFile(File ff) throws Exception{
		FileWriter fw = new FileWriter(ff);
        BufferedWriter bw = new BufferedWriter(fw);
        bw.write("NCOLS\t" + getNY() + "\n");
        bw.write("NROWS\t" + getNX() + "\n");
        bw.write("XLLCORNER\t" + getXll() + "\n");
        bw.write("YLLCORNER\t" + getYll() + "\n");
        bw.write("CELLSIZE\t" + getWindowCellSize(0, 0) + "\n");
        bw.write("NODATA_VALUE\t" + getNoDataValue() + "\n");
        
        for(int kk = 0; kk < layer.length; kk++){
        	for (int jj = 0; jj < layer[kk].length; jj++){
        		bw.write(Float.toString((float)layer[kk][jj]));
        		bw.write("\t");
        	}
        	bw.write("\n");
        }
        
        
        bw.close();
        fw.close();
	}
	
	public void initFromFile(String fileData, File localDir){
		try{
			String iccFileName = fileData;
			File file = new File(iccFileName);
			
			file = new File(localDir, file.getName());
			if (!file.exists()) {
					throw new FileNotFoundException(file.getAbsolutePath() + " not found!");
			}
			FileReader fr = new FileReader(file);
			BufferedReader br = new BufferedReader(fr);
			
			//las 6 primeras linias del fichero nos dans la forma
			StringTokenizer stk;
			final double[] tmpValues = new double[6];
			if (br.ready())
			for (int ii = 0; ii < 6; ii++){
				stk = new StringTokenizer(br.readLine());
				stk.nextToken();
				tmpValues[ii] = Double.parseDouble(stk.nextToken());
			}
			
			//assing constants of layer
			iNJ = (int)tmpValues[0];
			iNI = (int)tmpValues[1];
			xll = tmpValues[2];
			yll = tmpValues[3];
			cellSize = (int) tmpValues[4];
			
			if (cellSize == 15) layerType = RasterLayer.ICC;
			else if (cellSize == 30) layerType = RasterLayer.ICC30;
			_2DX = cellSize * 2.;
	        _6DX = cellSize * 6.;
			
			no_data = tmpValues[5];
			
			layer = new double[iNI][iNJ];
			
			//reading matrix of data
			int ii = 0, jj = 0;
			
			while( br.ready()) {
				stk = new StringTokenizer(br.readLine());
				
				while (stk.hasMoreTokens()){
					layer[ii][jj] = Double.parseDouble(stk.nextToken());
					jj++;
					
				}
				
				if (jj >= iNJ){
					jj=0;
					ii++;
				}
			}
			
			
			br.close();
			fr.close();
			
		}catch(Exception e){
			e.printStackTrace();
		}
	}
	
	/**
	 * This method allows to initialize the raster object from
	 * another raster object in order to keep only the form and not the 
	 * content
	 * @param other other Raster Layer
	 */
	public void initFromLayer(RasterLayer other){
		this.setLayer(other.getLayer()); //init iNX, iNY, _2DX, _6DX and layer
		this.assignNoData(); // set layer to no data values by default NO_DATA_VALUE
		this.setUTMOrig(other.getUTMOrig()); // set xll and yll
		this.cellSize = (int) other.getWindowCellSize((int)(iNI/2), (int)(iNJ/2));
	}
	
	public void setUTMOrig(double[] orig){
		this.xll = (float)orig[0];
		this.yll = (float)orig[1];
	}
	
	public String rasterType(){
		return this.layerType;
	}
	
	public double[] getUTMOrig(){
		double[] tmp = new double[2];
		tmp[0] = xll;
		tmp[1] = yll;
		
		return tmp;
	}
	
	public void setDataAtLatLon(double lat, double lon, double data){
		int iijj[] = get_iijj(lat, lon, GeoUTM.ETRS89, 31);
		layer[iijj[0]][iijj[1]] = data;
	}
	
	public void setDataAtXY_UTM(double xUTM, double yUTM, float data){
		int iijj[] = get_iijj(xUTM, yUTM);
		layer[iijj[0]][iijj[1]] = data;
	}
	
	public int[] get_iijj(double xUTM, double yUTM){
		int[] coords = new int[2];
		
		
//		coords[0] = iNI - (int)((yUTM-yll)/cellSize + 1/2);
//		coords[1] = (int)((xUTM-xll)/cellSize + 1/2);
//		
		coords[0] = iNI - (int)((yUTM-yll)/cellSize) + 1;
		coords[1] = (int)((xUTM-xll)/cellSize);
		
		//here the X coordinate is not good!!! always under zero!!
		
		if (coords[0] < 0) coords[0] = 0;
		if (coords[1] < 0) coords[1] = 0;
		if (coords[0] >= iNI) coords[0] = iNI - 1;
		if (coords[1] >= iNJ) coords[1] = iNJ - 1;
		
		return coords;
	}
	
	public double[] get_latlon(int ii, int jj, int type_ellipsoide, int zoneUTM){
		double[] coords = new double[2];
		double[] xyUTM = new double[2];
		
//		xyUTM[0] = jj * cellSize + xll;
//		xyUTM[1] = (iNI -1 - ii) * cellSize + yll;
		
		//this is the center of grid cell and not the corner
		xyUTM[0] = jj * cellSize + xll + cellSize/2.0;
		xyUTM[1] = (iNI + 1 - ii) * cellSize + yll + cellSize/2.0;
		
		GeoUTM gu = new GeoUTM(type_ellipsoide);
		
		coords = gu.passToGeographic(xyUTM[0], xyUTM[1], 31, false, false);
		
		//coords are in radiant, we need to pass it to degrees
		coords[0] *= RAD_TO_DEG;
		coords[1] *= RAD_TO_DEG;
		
		return coords;
	}
	
	private int[] get_iijj(double lat, double lon, int type_ellipsoide, int zoneUTM){
		int[] coords = new int[2];
		
		GeoUTM gu = new GeoUTM(type_ellipsoide);
		double[] xyUTM = gu.passToUTM(lat, lon, zoneUTM);
		
//		coords[0] = iNI - (int)((xyUTM[1]-yll)/cellSize + 1/2);
//		coords[1] = (int)((xyUTM[0]-xll)/cellSize + 1/2);
		
		coords[0] = iNI - (int)((xyUTM[1]-yll)/cellSize) + 1;
		coords[1] = (int)((xyUTM[0]-xll)/cellSize);
		
		if (coords[0] < 0) coords[0] = 0;
		if (coords[1] < 0) coords[1] = 0;
		if (coords[0] >= iNI) coords[0] = iNI - 1;
		if (coords[1] >= iNJ) coords[1] = iNJ - 1;
		
		return coords;
	}
	
	public String getLayerType(){
		return layerType;
	}
	/**
	 * Initcialitza el buffer
	 * @param b
	 */
	public void setLayer(double[][] bb){
		iNI = bb.length;
	    iNJ = bb[0].length;
		layer = new double[iNI][iNJ];
		for(int ii=0; ii<iNI; ii++)
			for(int jj=0; jj<iNJ; jj++)
				layer[ii][jj] = bb[ii][jj];
       
        _2DX = this.getWindowCellSize( (int)iNI/2, (int)iNJ ) * 2.;
        _6DX = this.getWindowCellSize( (int)iNI/2, (int)iNJ ) * 6.;
	}
	
	public double[][] getLayer(){
		return layer;
	}
	
	public void setLatMaxSRTM( float lat){
		fLatMax = lat;
	}
	
    public void setLonMinSRTM( float lat){
		fLonMin = lat;
	}
	
	public void multiply(double value){
		for (int ii=0; ii<iNI; ii++)
			for(int jj=0; jj<iNJ; jj++)
				if( !isNoDataValue(layer[ii][jj]) )
					layer[ii][jj] *= value;
	}
	
	
	public void assign(double value){
		for (int ii=0; ii<iNI; ii++)
			for(int jj=0; jj<iNJ; jj++)
				layer[ii][jj] = value;
	}
	
	public int getCellValueAsInt(int xx, int yy){
		return (int)layer[xx][yy];
	}
	public void setCellSize(int arcseconds){
		cellSize = arcseconds;
	}
	
	public void  setNoDataValue(double value){
		no_data = value;
	}
	
	public void assignNoData(){
		for (int ii=0; ii<iNI; ii++)
			for(int jj=0; jj<iNJ; jj++)
				layer[ii][jj] = no_data;
	}
	public void setNoDataAtCell(int xx, int yy){
		if (xx>=0 && xx<iNI && yy>=0 && yy<iNJ)
			layer[xx][yy] = no_data;
			
	}
	
	public double getDistToNeighborInDir(int ii){
		double distance;
		distance = getWindowCellSize((int)(iNI/2), (int)(iNJ/2));
		
		if (ii%2 != 0) //ii=1, 3, 5 o 7
			distance *= Math.sqrt(2);
		return distance;
	}
	/**
	 * This method allow to 
	 * @param xx
	 * @param yy
	 * @return
	 */
	public double getWindowCellSize(int xx, int yy){
		double size = 0;
		// en este caso son 3'' y la distancia depende de la latitud
		if (layerType.equals("SRTM")){
			double lat1, lat2;
			lat1 = (int)this.fLatMax  - yy/1200;
			lat2 = lat1 - 1/1200;
			
//			Haversine
//			formula: 	a = sin²(Δφ/2) + cos(φ1).cos(φ2).sin²(Δλ/2)
//			c = 2.atan2(√a, √(1−a))
//			d = R.c
//			where 	φ is latitude, λ is longitude, R is earth’s radius (mean radius = 6,371km)
//			  	note that angles need to be in radians to pass to trig functions!
			
			//calculo del radio equivalente
			double ex2 = 2 * APLA - APLA * APLA; //excentricidad al quadrado
			double R_equiv = R_EQU / Math.sqrt(1-ex2*Math.pow(Math.sin(lat1), 2));
			
			double dLat = DEG_TO_RAD * 1 / 1200;
			double dLon = DEG_TO_RAD * 1 / 1200;
			lat1 *= DEG_TO_RAD;
			lat2 *= DEG_TO_RAD;

			double a = Math.sin(dLat/2) * Math.sin(dLat/2) +
			        Math.sin(dLon/2) * Math.sin(dLon/2) * Math.cos(lat1) * Math.cos(lat2); 
			double c = 2 * Math.atan2(Math.sqrt(a), Math.sqrt(1-a)); 
			double d = R_equiv * c;
			
			size = d / Math.sqrt(2);
			
		}
		// en este caso la medida ICC es de 15m ou 30m
		else if (layerType.equals(RasterLayer.ICC) || layerType.equals(RasterLayer.ICC30))
			size = cellSize;
		return size;
	}
	public boolean isNoDataValue(double val){
		if (val == no_data)
			return true;
		return false;
	}
	
	public double getCellValueAsDouble(int xx, int yy){
            
        if( xx >= 0 && xx < iNI && yy >= 0 && yy < iNJ)
              return layer[xx][yy];
        return no_data;
	}
	
	public void setCellValue(int xx, int yy, double value){
		if( xx >= 0 && xx < iNI && yy >= 0 && yy < iNJ)
			layer[xx][yy] = value;
	}
	
	public int getNX(){
		return iNI;
	}
	
	public int getNY(){
		return iNJ;
	}
	
//////////////////////////////Additional methods for DEM analysis//////
// from The JUMP Pilot Project Sourceforge  /core/trunk/src/org/openjum/core/rasterimage/sextante

	private boolean getSubMatrix3x3(int x, int y, double SubMatrix[]){

		int	i;
		int iDir;
		double	z, z2;
		
		z = getCellValueAsDouble(x, y);
		
		if(isNoDataValue(z)){
			return false;
		}
		else{
		//SubMatrix[4]	= 0.0;
			for(i=0; i<4; i++){

				iDir = 2 * i;
				z2 = getCellValueAsDouble(x + m_iOffsetX[iDir], y + m_iOffsetY[iDir]);
				if( !isNoDataValue(z2)){
					SubMatrix[i]	=  z2 - z;
				}
				else{
					z2 = getCellValueAsDouble(x + m_iOffsetX[(iDir + 4) % 8], y + m_iOffsetY[(iDir  + 4) % 8]);
					if( !isNoDataValue(z2)){
						SubMatrix[i]	= z - z2;
					}
					else{
						SubMatrix[i]	= 0.0;
					}
				}
			}
		
			return true;
		}

	}

	public double getSlope(int x, int y){

		double	zm[], G, H;

		zm = new double[4];

		if( getSubMatrix3x3(x, y, zm) ){
			G	=  (zm[0] - zm[2]) / _2DX;
			H	=  (zm[1] - zm[3]) / _2DX;
			return Math.atan(Math.sqrt(G*G + H*H));
		}
		else{
			return no_data;
		}
	}

	public double getAspect(int x, int y){

		double	zm[], G, H, dAspect;

		zm = new double[4];

		if( getSubMatrix3x3(x, y, zm) ){
			G	=  (zm[0] - zm[2]) / _2DX;
			H	=  (zm[1] - zm[3]) / _2DX;
			if( G != 0.0 ){
				dAspect = DEG_180_IN_RAD + Math.atan2(H, G);
			}
			else{
				dAspect = H > 0.0 ? DEG_270_IN_RAD : (H < 0.0 ? DEG_90_IN_RAD : -1.0);
			}
			return dAspect;
		}
		else{
			return no_data;
		}
	}

	public static double getUnitDistToNeighborInDir(int iDir){

		return( (iDir % 2 != 0) ? Math.sqrt(2.0)  : 1.0 );

	}
	
	
	public int getDirToNextDownslopeCell(int x, int y){

		return getDirToNextDownslopeCell(x, y, true);

	}

	public int getDirToNextDownslopeCell(int x, int y, boolean bForceDirToNoDataCell){

		int i, iDir;
		double	z, z2, dSlope, dMaxSlope;

		z = getCellValueAsDouble(x, y);

		if(isNoDataValue(z)){
			return -1;
		}

		dMaxSlope = 0.0;
		for(iDir=-1, i=0; i<8; i++){
			z2 = getCellValueAsDouble(x + m_iOffsetX[i], y + m_iOffsetY[i]);
			if(isNoDataValue(z2)){
				if (bForceDirToNoDataCell){
					return i;
				}
				else{
					return -1;
				}
			}
			else{
				dSlope	= (z - z2) / getDistToNeighborInDir(i);
				if( dSlope > dMaxSlope ){
					iDir = i;
					dMaxSlope = dSlope;
				}
			}
		}

		return iDir;

	}
        public double getCellValueInLayerCoords(int xx, int yy, int band){
            return layer[xx][yy];
        }
        
        public double getXMin(){
            return fLonMin;
        }
        public double getYMax(){
            return fLatMax;
        }
        
        public double getValueAt(double xPosition, double yPosition, int band){

		int	x, y;
		double	dx, dy;
		double dValue;
                
		x = (int) Math.floor(xPosition = (xPosition - getXMin()) / 
                        getWindowCellSize((int)(iNI/2), (int)(iNJ/2)));
		y = (int) Math.floor(yPosition = (getYMax() - yPosition ) / 
                        getWindowCellSize((int)(iNI/2), (int)(iNJ/2)));

		dValue = getCellValueInLayerCoords(x,y,band);

		if( !isNoDataValue(dValue) ){

			dx = xPosition - x;
			dy = yPosition - y;

			switch( m_iMethod ){
			case INTERPOLATION_NearestNeighbour:
				dValue = getValueNearestNeighbour (x, y, dx, dy, band);
				break;

			case INTERPOLATION_Bilinear:
				dValue	= getValueBiLinear (x, y, dx, dy, band);
				break;

			case INTERPOLATION_InverseDistance:
				dValue	= getValueInverseDistance(x, y, dx, dy, band);
				break;

			case INTERPOLATION_BicubicSpline:
				dValue	= getValueBiCubicSpline (x, y, dx, dy, band);
				break;

			case INTERPOLATION_BSpline:
				dValue	= getValueBSpline (x, y, dx, dy, band);
				break;
			}
		}
		else{
			dValue = no_data;
		}

		return dValue;
	}

	private double getValueNearestNeighbour(int x, int y, double dx, double dy, int band){

		x	+= (int)(0.5 + dx);
		y	+= (int)(0.5 + dy);

		return getCellValueInLayerCoords(x, y, band);

	}

	private  double getValueBiLinear(int x, int y, double dx, double dy, int band){

		double	z = 0.0, n = 0.0, d;
		double dValue;

		dValue = getCellValueInLayerCoords(x, y, band);
		if (!isNoDataValue(dValue)){
			 d = (1.0 - dx) * (1.0 - dy);
			 z += d * dValue;
			 n += d;
		}

		dValue = getCellValueInLayerCoords(x + 1, y, band);
		if (!isNoDataValue(dValue)){
			 d = (dx) * (1.0 - dy);
			 z += d * dValue;
			 n += d;
		}

		dValue = getCellValueInLayerCoords(x, y + 1, band);
		if (!isNoDataValue(dValue)){
			 d = (1.0 - dx) * (dy);
			 z += d * dValue;
			 n += d;
		}

		dValue = getCellValueInLayerCoords(x + 1, y + 1, band);
		if (!isNoDataValue(dValue)){
			 d = (dx) * (dy);
			 z += d * dValue;
			 n += d;
		}

		if( n > 0.0 ){
			return( z / n );
		}

		return( no_data );
	}

	private double getValueInverseDistance(int x, int y, double dx, double dy, int band){

		double	z = 0.0, n = 0.0, d;
		double dValue;

		if( dx > 0.0 || dy > 0.0 ){

			dValue = getCellValueInLayerCoords(x, y, band);
			if (!isNoDataValue(dValue)){
				d = 1.0 / Math.sqrt(dx*dx + dy*dy);
				z += d * dValue;
				n += d;
			}

			dValue = getCellValueInLayerCoords(x + 1, y, band);
			if (!isNoDataValue(dValue)){
				d = 1.0 / Math.sqrt((1.0-dx)*(1.0-dx) + dy*dy);
				z += d * dValue;
				n += d;
			}

			dValue = getCellValueInLayerCoords(x, y + 1, band);
			if (!isNoDataValue(dValue)){
				d = 1.0 / Math.sqrt(dx*dx + (1.0-dy)*(1.0-dy));
				z += d * dValue;
				n += d;
			}

			dValue = getCellValueInLayerCoords(x + 1, y + 1, band);
			if (!isNoDataValue(dValue)){
				d = 1.0 / Math.sqrt((1.0-dx)*(1.0-dx) + (1.0-dy)*(1.0-dy));
				z += d * dValue;
				n += d;
			}

			if( n > 0.0 )
			{
				return( z / n );
			}
		}
		else{
			return getCellValueInLayerCoords(x, y, band) ;
		}

		return( no_data);
	}

	private double getValueBiCubicSpline(int x, int y, double dx, double dy, int band){

		int		i;
		double	a0, a2, a3, b1, b2, b3, c[], z_xy[][];

		c = new double[4];
		z_xy = new double[4][4];

		if( get4x4Submatrix(x, y, z_xy, band) ){

			for(i=0; i<4; i++){
				a0		= z_xy[0][i] - z_xy[1][i];
				a2		= z_xy[2][i] - z_xy[1][i];
				a3		= z_xy[3][i] - z_xy[1][i];

				b1		= -a0 / 3.0 + a2       - a3 / 6.0;
				b2		=  a0 / 2.0 + a2 / 2.0;
				b3		= -a0 / 6.0 - a2 / 2.0 + a3 / 6.0;

				c[i]	= z_xy[1][i] + b1 * dx + b2 * dx*dx + b3 * dx*dx*dx;
			}

			a0		= c[0] - c[1];
			a2		= c[2] - c[1];
			a3		= c[3] - c[1];

			b1		= -a0 / 3.0 + a2       - a3 / 6.0;
			b2		=  a0 / 2.0 + a2 / 2.0;
			b3		= -a0 / 6.0 - a2 / 2.0 + a3 / 6.0;

			return( c[1] + b1 * dy + b2 * dy*dy + b3 * dy*dy*dy );
		}

		return( getValueBiLinear(x, y, dx, dy, band) );
	}

	private double getValueBSpline(int x, int y, double dx, double dy, int band){

		int		i, ix, iy;
		double	z, px, py, Rx[], Ry[], z_xy[][];

		Rx = new double[4];
		Ry = new double[4];
		z_xy = new double [4][4];

		if( get4x4Submatrix(x, y, z_xy, band) ){
			for(i=0, px=-1.0-dx, py=-1.0-dy; i<4; i++, px++, py++){
				Rx[i]	= 0.0;
				Ry[i]	= 0.0;

				if( (z = px + 2.0) > 0.0 )
					Rx[i]	+=        z*z*z;
				if( (z = px + 1.0) > 0.0 )
					Rx[i]	+= -4.0 * z*z*z;
				if( (z = px + 0.0) > 0.0 )
					Rx[i]	+=  6.0 * z*z*z;
				if( (z = px - 1.0) > 0.0 )
					Rx[i]	+= -4.0 * z*z*z;
				if( (z = py + 2.0) > 0.0 )
					Ry[i]	+=        z*z*z;
				if( (z = py + 1.0) > 0.0 )
					Ry[i]	+= -4.0 * z*z*z;
				if( (z = py + 0.0) > 0.0 )
					Ry[i]	+=  6.0 * z*z*z;
				if( (z = py - 1.0) > 0.0 )
					Ry[i]	+= -4.0 * z*z*z;

				Rx[i]	/= 6.0;
				Ry[i]	/= 6.0;
			}

			for(iy=0, z=0.0; iy<4; iy++){
				for(ix=0; ix<4; ix++){
					z	+= z_xy[ix][iy] * Rx[ix] * Ry[iy];
				}
			}

			return( z );
		}

		return( getValueBiLinear(x, y, dx, dy, band) );
	}

	private boolean get4x4Submatrix(int x, int y, double z_xy[][], int band){

		int	ix, iy, px, py;
		double dValue;

		for(iy=0, py=y-1; iy<4; iy++, py++){
			for(ix=0, px=x-1; ix<4; ix++, px++){
				dValue = getCellValueInLayerCoords(px, py , band);
				if (isNoDataValue(dValue)){
					return false;
				}
				else{
					z_xy[ix][iy] = dValue;
				}
			}
		}

		return( true );
	}

	public void setInterpolationMethod(int iMethod){

		m_iMethod = iMethod;

	}
}
